#include <stdio.h>
#include <math.h>


void burst();
void overunder();
void packets();
double stats(); 

main(argc, argv) 
  int argc;
  char *argv[];
{
  char *fname;
  float x;
  double stddev;

  if (argc < 2) {
     printf("Usage: stddev [datafile]\n");
     exit();
  } else {
     fname = argv[1];
  }
  printf("\n\nStatistics file for: %s\n", fname);
  x = 0;
  stddev = stats(fname);  
  while ((x < 3) && (x > -3)) {
    overunder(fname, stddev*x); 
    x -= 0.1;
  }

/*  burst(fname); */ 
/*   packets(fname); */ 
}

#define FPS 30         /* Frame Rate -- frames per second */
#define CELLSIZE 53      /* Packet Size -- bytes per packet */

/*-----------------------------------------*/

void overunder(fname, stddev) 
  char *fname;
  double stddev;
{
  int cnt;
  FILE *fd;
  int maxOflow, maxUflow;
  int maxOflowSlots, maxUflowSlots;
  int fnum, fsize;
  int runtotal, frametotal;
  int rem, delta;
  int Ls;        /* total stream size in bits */
  int Nf;        /* total number of frames in stream */
  double Ts;     /* total amount of time to deliver stream */
  int Rb;        /* bit rate required */
  int OflowSlots, UflowSlots;
  int i, buffsize, Bbits;

  fd = fopen(fname, "r");
  Ls = Nf = Rb = 0;
  Ts = 0;
  while (!feof(fd)) {
    fscanf(fd, "%d %d\n", &fnum, &fsize);
    Ls += fsize;
    Nf++;
  }
  fseek(fd,0,0);
  
  Ts = Nf/(FPS*1.0);
  Rb = irint(floor(Ls/Ts)) + irint(floor(stddev));

printf("Ls %d, Nf %d, Ts %e, Rb %d, dev %e\n", Ls, Nf, Ts, Rb, stddev);

  maxOflow = maxUflow = 0;
  maxOflowSlots = maxUflowSlots = 0;
  OflowSlots = UflowSlots = 0;
  runtotal = frametotal = 0;
  rem = i = delta = 0;

  while (!feof(fd)) {
    fscanf(fd, "%d %d\n", &fnum, &fsize);
    runtotal++;
    frametotal += fsize;
   
#if DEBUG
printf("Fno %d, Fsize %d, Fnumt %d, Fsizet %d\n", 
       fnum, fsize, runtotal, frametotal);
#endif
    
    if (frametotal >= Rb) {            /* ONE TIME SLOT HAS ELAPSED */
      rem = frametotal - Rb;
      if (rem > 0) {
	runtotal--;
      }

      delta += runtotal - FPS;
      if (delta < maxUflow) {          /* delta < 0 => underflow */
	maxUflow = delta;              /* store largest underflow */
      }
      if (delta > maxOflow) {          /* delta > 0 => overflow */
	maxOflow = delta;              /* store largest overflow */
      }

      if (runtotal > FPS){
	OflowSlots++;
      }
      if (runtotal < FPS){
	UflowSlots++;
      }
      cnt = UflowSlots - OflowSlots;
      if (cnt > maxUflowSlots) {
	maxUflowSlots = cnt;
      }
      cnt = OflowSlots - UflowSlots;
      if (cnt > maxOflowSlots) {
	maxOflowSlots = cnt;
      }
#if DEBUG
printf("\nofslots %d, ufslots %d, maxuflow %d, maxoflow %d\n\n",
       maxOflowSlots, maxUflowSlots, maxUflow, maxOflow);
#endif

      frametotal = rem;
      if (rem > 0) runtotal = 1;
      else runtotal = 0;
    }
  }

  printf("\n");
  printf("OVERFLOW/UNDERFLOW-SLOT DATA: (rate = %d)\n", Rb);
  printf("Overflow Slots       : %d\n", maxOflowSlots);
  printf("Underflow Slots      : %d\n", maxUflowSlots);
  printf("Max. Underflow Frames: %d\n", maxUflow);
  printf("Max. Overflow Frames : %d\n", maxOflow);
  printf("\n");

  fseek(fd,0,0);
  buffsize = 0;
  for(i=0; i > maxUflow; i--) {
    fscanf(fd, "%d %d\n", &fnum, &fsize);
    buffsize += fsize;
  }
  printf("Underflow Buffer req.: %d (bits), %f (bytes)\n", buffsize,
         buffsize/8.0);
  Bbits = GetOflowBits(fd, Rb, maxUflow);
  printf("Overflow Buffer req. : %d (bits), %f (bytes)\n", Bbits,
         Bbits/8.0);
  printf("\n\n");
  fclose(fd);
}

/*-----------------------------------------*/

int GetOflowBits(fd, Rb, Uflow)
  FILE *fd;
  int Rb;        /* bit rate required */
  int Uflow;     /* max underflow frames */
{
  struct Queue{
    int fsize;
    int fno;
    struct Queue *next; 
  };

  struct Queue *LowWater = NULL; 
  struct Queue *tmpPtr = NULL; 
  struct Queue *Consumed = NULL; 
  struct Queue *Current = NULL; 
  struct Queue *Qhead = NULL; 
  struct Queue *Qptr = NULL; 
  int fno, fsize;
  int Framecount;
  int FrameSizeCount = 0;
  int rem, i;
  int OflowBits=0, maxOflow=0;
  int ConsumedFrames = 0;
  int LowWaterFrames = 0;
  int SlotNo = 0;

  fseek(fd,0,0);
  Framecount = 0;

  Uflow = Uflow * -1;
  if ((Uflow % FPS) > 0) {
    Uflow = (Uflow/FPS)*FPS + FPS;           /* ROUND to nearest slot */
  }

  i = 0;
  while (i < Uflow) {                        /* load underflow buffer */
     fscanf(fd, "%d %d\n", &fno, &fsize); 

#ifdef DEBUG
     if (fno%30 == 0) {
        SlotNo++;
        printf("%d %d %d %d\n", SlotNo, fno, 0, fno); /*####*/
     }
#endif

     Qptr = (struct Queue *)malloc(sizeof(struct Queue));
     if (Qptr == NULL) {
        printf("Malloc failure\n");
        exit();
     }
     Qptr->fsize = fsize;
     Qptr->fno = fno;
     Qptr->next = NULL;
     
     if (Qhead == NULL) {
        Qhead = Qptr;
        Consumed = Qptr;
        Current = Qptr;
     } else {
        Current->next = Qptr;
        Current = Qptr;
     }
     i++;
  }
  LowWater = Current;                       /* set low water mark */
  LowWaterFrames = Current->fno;

  while (!feof(fd)) {                       /* begin display */ 
     fscanf(fd, "%d %d\n", &fno, &fsize); 
     Framecount++;
     FrameSizeCount += fsize;
 
     Qptr = (struct Queue *)malloc(sizeof(struct Queue));
     Qptr->fsize = fsize;
     Qptr->fno = fno;
     Qptr->next = NULL;
     
     Current->next = Qptr;
     Current = Qptr;

     if (Consumed == NULL) Consumed = Qptr;

     if (FrameSizeCount > Rb) {      /* ONE TIME SLOT HAS ELAPSED */
        rem = FrameSizeCount - Rb;
        i = 0; 
        while (i < FPS) {
           ConsumedFrames++;
           if (Consumed == NULL) {
              printf("WARNING!!!   UNDERFLOW Condition\n");
              exit();
           } 
           if (Consumed == LowWater) {      /* underflow buffer all used up */
              LowWater = Consumed->next;
              LowWaterFrames = ConsumedFrames;
           }
           Consumed = Consumed->next;
           i++;
        } 

#ifdef DEBUG
    printf("%d %d %d %d\n", SlotNo, LowWaterFrames, ConsumedFrames, Current->fno); /*###*/
#endif
        tmpPtr = LowWater;
        OflowBits = 0;
        while (tmpPtr != NULL) {
           OflowBits += tmpPtr->fsize;
           tmpPtr = tmpPtr->next;
        }  
        if (OflowBits > maxOflow) {
           maxOflow = OflowBits;
        }
        FrameSizeCount = rem;     
     }
  }
  return(maxOflow); 
}

/*-----------------------------------------*/

void burst(fname) 
   char *fname;
{
   FILE *fd;
   int size, index;
   double avesize;
   int floorave;
   int burstlen, newburst, frameburstlen, totalburstlen, lastburst;
   int totialen, ialen, iacnt, newia;
   float aveframeburstlen, aveburstlen, aveialen;  
   int maxframeburstlen, totalsize, maxburstlen, burstcount; 
   int totalframeburstlen;
   int maxburstseq, burstseq;

   totalsize = 0;
   fd = fopen(fname,"r");
   while (!feof(fd)) {
      fscanf(fd, "%d %d\n", &index, &size);
      totalsize += size;
   }

   avesize = totalsize/index;
   floorave = irint(floor(avesize));
 
   if (fseek(fd, 0, 0) < 0) printf("ERROR SEEK\n");  

   burstlen = 0;
   newburst = 1;
   frameburstlen = totalburstlen = totalframeburstlen = 0;
   maxframeburstlen = maxburstlen = lastburst = 0;
   burstcount = 0;
   newia = 1;
   totialen = iacnt = ialen = 0;
   maxburstseq = 0;
   burstseq = 0;     /* assigns +1 to bursts and -1 (mod -1) to non-burst frames */
		     /*   we store the largest burst sequence over all frames */

   while (!feof(fd)) {
      fscanf(fd, "%d %d\n", &index, &size); 
      if (size > floorave) {
	  burstlen++;             /* count no. of frames in burst */
	  frameburstlen += size;  /* count size of frames in burst */ 
	  if (newburst) {
	      burstcount++;       /* count no. of bursts */ 
	      newburst = 0;
	  }

	  burstseq += 1;
  	  if (burstseq > maxburstseq) {
	      maxburstseq = burstseq;
          }

	  newia = 1;                       
	  totialen += ialen;     /* sum Total no. of frames between bursts */ 
	  ialen = 0;
      } else {
	  burstseq += -1;
	  if (burstseq < 0) burstseq = -1;

	  totalburstlen += burstlen;      /* sum total no. of frames for all bursts */
	  totalframeburstlen += frameburstlen;  /* sum of total frame bytes for all bursts */
	  if (burstlen > maxburstlen) {
 	      maxburstlen = burstlen;
          }

	  if (frameburstlen > maxframeburstlen) {    
             maxframeburstlen = frameburstlen;   /* max total no. of frame bytes in a burst */
	  }

	  if (newia) {
             iacnt++;       /* count no. slots between bursts  */
	     newia = 0;     /*    should be between 1 or 2 of burst count */ 
  	  }

   	  ialen++;          /* count no. of frames between bursts */ 
	  newburst = 1;
	  burstlen = 0; 
	  frameburstlen = 0;
      }
   } 

   aveburstlen = totalburstlen/burstcount;            /* ave. no. of frames per burst */
   aveframeburstlen = totalframeburstlen/burstcount;  /* ave. no. of bytes per burst */
   aveialen = totialen/iacnt;                         /* ave. no. of frames between bursts */

   fclose(fd); 

   printf("WORKLOAD BURST CHARACTERISTICS:\n");
   printf("Ave. Burst Length            : %e (frames)\n", aveburstlen);
   printf("Ave. No. of Bits per Burst   : %e\n", aveframeburstlen);
   printf("Max. No. of Frames in a Burst: %d\n", maxburstlen); 
   printf("Max. No. of Bits in a Burst  : %d\n", maxframeburstlen); 
   printf("Largest Sequence of Bursts   : %d (frames)\n", maxburstseq);
   printf("Ave. Interarrival Length     : %e (frames)\n", aveialen); 
   printf("\n"); 
}

/*------------------------------------------------------*/
void packets(fname) 
   char *fname;
{
   FILE *fd, *fd2;
   int fno, fsize;
   char fname2[30];
   int numcells;
   double maxcells;
   double dnumcells, cellcount, framecount;
   double cellcount_sqr;
   double meancells, variance, stddev;
   double padding, meanpad, padvar, padcnt;
   double maxpad, padstddev, pad_sq;
 
   fd = fopen(fname, "r");
   strcpy(fname2, strtok(fname, "."));
   strcat(fname2,".cells.dat"); 
   fd2 = fopen(fname2, "w");

   framecount = cellcount = maxcells = 0.0;
   padding = meanpad = padvar = padcnt = 0.0;
   padstddev = maxpad = pad_sq = 0.0;
   cellcount_sqr = 0.0;
   while (!feof(fd)) {
      fscanf(fd, "%d %d\n", &fno, &fsize);
      numcells = 0;
      dnumcells = 0.0;
      while (fsize > 0) {
         if (fsize < CELLSIZE) {            /* next cell will be padded */
             padding = CELLSIZE - fsize;
             if (padding > maxpad) maxpad = padding;
             meanpad += padding;            /* use meanpad as an accumulator */
             padcnt += 1.0;
             pad_sq += pow(padding,2.0);
         }
         fsize -= CELLSIZE; 
         dnumcells += 1.0;  
         numcells++;  
      }
      fprintf(fd2,"%d %d\n", fno, numcells); 
      framecount += 1.0;
      cellcount += dnumcells; 
      cellcount_sqr += pow(dnumcells,2.0); 
      if (dnumcells > maxcells) maxcells = dnumcells;
   }
   fclose(fd);
   fclose(fd2);

   meancells = cellcount/framecount; 
   variance = (cellcount_sqr/framecount) - pow(meancells,2.0);
   stddev = sqrt(variance);

   meanpad = meanpad/padcnt;
   padvar = (pad_sq/padcnt) - pow(meanpad,2.0); 
   padstddev = sqrt(padvar);

   printf("CELL-PER-FRAME STATISTICS:\n");
   printf("Max. Number of Cells/Frame : %e\n", maxcells);
   printf("Ave. Number of Cells/Frame : %e\n", meancells);
   printf("Cells/Frame Variance       : %e\n", variance); 
   printf("Cells/Frame Std. Deviation : %e\n", stddev); 
   printf("\n");
   printf("FRAME PADDING STATISTICS:\n"); 
   printf("Max. Padding per Frame : %e\n", maxpad);
   printf("Ave. Padding per Frame : %e\n", meanpad);
   printf("Padding Variance       : %e\n", padvar); 
   printf("Padding Std. Deviation : %e\n", padstddev); 
   printf("\n\n"); 
}

/*------------------------------------------------------*/
double stats(fname) 
   char *fname;
{
   FILE *fd;
   int fno, fsize;
   double maxframesize;
   double framesize, framecount, totframesize;
   double framesize_sqr;
   double meanframesize, variance, stddev;

   fd = fopen(fname, "r");

   totframesize = maxframesize = 0.0;
   framesize_sqr = framecount = 0.0;
   while (!feof(fd)) {
      fscanf(fd, "%d %d\n", &fno, &fsize);
      framecount += 1.0;
      totframesize += fsize;
      framesize = fsize;
      framesize_sqr += pow(framesize,2.0);
      if (framesize > maxframesize) maxframesize = framesize;
   }
   fclose(fd);

   meanframesize = totframesize/framecount;
   variance = (framesize_sqr/framecount) - pow(meanframesize,2.0);
   stddev = sqrt(variance);

   printf("FRAME SIZE STATISTICS:\n"); 
   printf("Max. Frame Size           : %e\n", maxframesize);
   printf("Ave. Frame Size           : %e\n", meanframesize);
   printf("Frame Size Variance       : %e\n", variance);
   printf("Frame Size Std. Deviation : %e\n", stddev);
   printf("\n");
   
   return(stddev);
}
